(* file: movies.ml
 * author: Bob Muller, adapted from a Python program of Howard Straubing.
 *
 * CS1101 Computer Science I
 *
 * This is a problem set relating to working with dictionaries/maps. The map entries are strings:
 * names of actors and movies in which they appeared.
 *
 * To run:
 *
 * > make
 * > ./movies mdb.txt
 *)

(* Make an implementation of maps from Strings to anything. See
 *
 * http://caml.inria.fr/pub/docs/manual-ocaml/libref/Map.Make.html
 *
 * For a description of the various operations M.empty, M.mem, M.find, M.cardinal, M.filter, M.fold, etc.
 *)
module M = Map.Make(String)

(* Debugging switch.
 *)
let debug = false

(* PART 1 -- queries related to Film/Actors dictionary. *********************************************** *)

(* addFilmActor : Map.t -> string -> string -> Map.t
 *
 * The call (addFilmActor map film actor) returns the map resulting from adding the key/value 
 * film -> actor to map.
 *)
let addFilmActor map film actor =
  match M.mem film map with
  | true  -> let actors = M.find film map
             in
             (match List.mem actor actors with
              | true  -> map
              | false -> M.add film (actor :: actors) map)
               
  | false -> M.add film [actor] map

let makeFilmActorsMap filename =
  let inch = open_in filename in
  let rec repeat films actors =
    (try let film  = input_line inch in
         let actor = input_line inch
         in
         repeat (film :: films) (actor :: actors)
     with
       End_of_file -> close_in inch;
                      (films, actors)) in
  let (films, actors) = repeat [] []
  in
  List.fold_left2 addFilmActor M.empty films actors

(* The Film/Actors Dictionary.  Note that Sys.argv.(1) will be the name of the file containing
 * the actor/film database entries.
 *)
let faMap = makeFilmActorsMap (Sys.argv.(1))

(* 1. What is the length of the longest cast list?
 *)
let one () =
  let longest film actors maxLen =
    let _ = if debug then
              Printf.printf "One: %s has %d actors.\n" film (List.length actors)
            else ()
    in
    max (List.length actors) maxLen
  in
  M.fold longest faMap 0

let _ =
  let n = one ()
  in
  Printf.printf "One: Longest cast = %d.\n" n
            
(* 2. What is the film or films with the largest cast?
 *)
let two () =
  let longest film actors (longName, maxLen) =
    let _ = if debug then
              Printf.printf "% s has %d actors.\n" film (List.length actors)
            else ()
    in        
    if (List.length actors) > maxLen then
      (film, List.length actors)
    else
      (longName, maxLen)
  in
  let (answer, _) = M.fold longest faMap ("No Film", 0)
  in
  Printf.printf "\nTwo: %s had the most actors.\n" answer;
  answer

let _ = two () 

(* 3. How many films are in the database?
 *)
let three () = Printf.printf "\nThree: There are %d films.\n" (M.cardinal faMap)

let _ = three ()
              
(* 4. Answer questions 1 and 2 above for 'smallest cast'.
 *)
         
(* 5. List all the movies in which Owen Wilson appears. (Solve this
 *    using the Film/Actors dictionary faMap.  In the next step you'll see a simpler
 *    solution using a different dictionary.)
 *)
let five () =
  let owens film actors = List.mem "Owen Wilson" actors in
  let owenMap = M.filter owens faMap
  in
  let nOwens = M.cardinal owenMap
  in
  Printf.printf "\nFive: Owen Wilson appeared in %d films.\n" nOwens;
  nOwens

let _ = five ()

(* Some helper functions. *)
             
let rec intersection xs ys =
  match xs with
  | [] -> []
  | x :: xs -> let ans = intersection xs ys
               in
               (if List.mem x ys then (x :: ans) else ans)
    
let rec union xs ys =
  match xs with
  | [] -> ys
  | x :: xs -> let ans = union xs ys
               in
               (match List.mem x ans with
                | true  -> ans
                | false -> x :: ans)
                 
let rec printList xs =
  match xs with
  | [] -> ()
  | x :: xs -> Printf.printf "%s\n" x;
               printList xs
                         
(* 6. List all the actors who appeared in both 'Silver Linings Playbook'
 *    and 'American Hustle'.
 *)
let six () =
  let actors1 = M.find "Silver Linings Playbook" faMap in
  let actors2 = M.find "American Hustle" faMap in
  let _ = if debug then
            begin
              print_string "\n\n";
              printList actors1;
              print_string "\n";
              printList actors2
            end else () in
  let answer = intersection actors1 actors2
  in
  print_string "\nSix: Appeared in both Silver Linings and American Hustle:\n"; printList answer;
  answer

let _ = six ()

(* PART 2 -- queries related to Actor/Films dictionary. *********************************************** *)

(* addActorFilm : Map.t -> string -> string -> Map.t
 *
 * The call (addActorFilm map actor film) returns the map resulting from adding the key/value 
 * actor -> film to map.
 *)
let addActorFilm map film actor =
  match M.mem actor map with
  | true  -> let films = M.find actor map
             in
             (match List.mem film films with
              | true  -> map
              | false -> M.add actor (film :: films) map)
  | false -> M.add actor [film] map

let makeActorFilmsMap filename =
  let inch = open_in filename in
  let rec repeat films actors =
    (try let film  = input_line inch in
         let actor = input_line inch
         in
         repeat (film :: films) (actor :: actors)
     with
       End_of_file -> close_in inch;
                      (films, actors)) in
  let (films, actors) = repeat [] []
  in
  List.fold_left2 addActorFilm M.empty films actors

                  
(* The Actor/Films Dictionary.
 *)
let afMap = makeActorFilmsMap (Sys.argv.(1))

(* 7. List (again) all the movies in which Owen Wilson appears.
 *)
let seven () =
  let owenFilms = M.find "Owen Wilson" afMap
  in
  print_string "\nSeven: Owen Wilson Films:\n";
  printList owenFilms;
  owenFilms

let _ = seven ()

(* 8. How many actors are in the database?
 *)
let eight () =
  let nActors = M.cardinal afMap
  in
  Printf.printf "\nEight: there are %d actors in all.\n" nActors;
  nActors

let _ = eight ()

(* 9. Which actor (or actors) has been in the largest number of films
 *    in the database?  [By the way, because of the large number of
 *    animated films in the database, answers to questions like these
 *    tend to include hard-working voice actors whom you probably
 *    never heard of.]
 *)
let nine () =
  let longest actor films (longName, maxLen) =
    if (List.length films) > maxLen then
      (actor, List.length films)
    else
      (longName, maxLen)
  in
  let (answer, _) = M.fold longest afMap ("No Film", 0)
  in
  Printf.printf "\nNine: %s was in the most films.\n" answer;
  answer

let _ = nine ()
             
(* 10. What are all the films in which either Clint Eastwood or Morgan
 *     Freeman appeared?
 *)
let ten () =
  let ceFilms = M.find "Clint Eastwood" afMap in
  let mfFilms = M.find "Morgan Freeman" afMap in
  let both = union ceFilms mfFilms
  in
  print_string "\nTen: films with either Eastwood or Freeman=\n";
  printList both;
  both

let _ = ten ()

(* PART 3 -- queries related to collaborators. *********************************************** *)

(* A couple of helper functions. *)
            
let stringList2String sl =
  let rec loop sl =
    match sl with
    | [] -> ""
    | [s] -> s
    | s :: ss -> let ans = loop ss
                 in
                 (s ^ "; ") ^ ans
  in
  "[" ^ (loop sl) ^ "]"
     
let pdb db n =
  let dbbs = M.bindings db in
  let rec loop n dbbs =
    match (n = 0, dbbs) with
    | (true, _)  -> ()
    | (false, (s, sl) :: rest) -> 
       Printf.printf "\nkey=%s, val=%s" s (stringList2String sl);
       loop (n - 1) rest
    | _ -> failwith "This cannot happen"
  in
  loop n dbbs                   

let addActorCollaborators map actor collaborators = M.add actor collaborators map

let makeCollaboratorMap () =
  let actorFilms = M.bindings afMap in
  let acss =
    List.map
      (fun (actor, films) ->
       let collaborators =
         List.fold_left
           (@)
           []
           (List.map (fun film -> M.find film faMap) films)
       in
       (actor, collaborators))
      actorFilms in
  let accs' : (string * string list) list = List.map
                (fun (actor, colabs) ->
                 let colabs' = List.filter
                                 (fun actor' -> actor <> actor') colabs
                 in
                 (actor, colabs'))
                acss in
  let (actors, collaborators) = List.split accs'
  in
  List.fold_left2 addActorCollaborators M.empty actors collaborators

                  
(* The Collaborators Dictionary.
 *)
let coMap = makeCollaboratorMap ()

(* 11. Who is the actor with the most collaborators?  (It's not Kevin Bacon)
 *)
let eleven () = 
  let longest actor collaborators (longName, maxLen) =
    let _ = if debug then
              Printf.printf "% s has %d collaborators.\n" actor (List.length collaborators)
            else
              ()
    in
    if (List.length collaborators) > maxLen then
      (actor, List.length collaborators)
    else
      (longName, maxLen)
  in
  let (answer, _) = M.fold longest coMap ("No Actor", 0)
  in
  Printf.printf "\nEleven: %s had the most collaborators.\n" answer;
  answer

let _ = eleven () 
    
(* 12. Is Kate Winslet a collaborator of Cate Blanchett?  
 *)
let twelve () =
  let collaborators = M.find "Cate Blanchett" coMap
  in
  match List.mem "Kate Winslet" collaborators with
  | true  -> Printf.printf "\nTwelve: Yes, Kate Winslet is a collaborator of C. Blanchett.\n"
  | false -> Printf.printf "\nTwelve: No, Kate Winslet isn't a collaborator of C. Blanchett.\n"

let _ = twelve ()
               
(* 13. Is Kate Winslet a collaborator of a collaborator of Cate Blanchett?  
 *)
let thirteen () =
  let collaborators = M.find "Cate Blanchett" coMap 
  in
  List.exists (fun c -> List.mem "Kate Winslet" (M.find c coMap)) collaborators

let _ =
  match thirteen() with
  | true  -> Printf.printf "\nThirteen: Yes, Kate Winslet collaborated with a collaborator of Cate Blanchett.\n"
  | false -> Printf.printf "\nThirteen: No, Kate Winslet has not collaborated with a collaborator of Cate Blanchett.\n"


(* PART 4 -- queries related to collaboration graphs. *********************************************** *)

(* pseudo-code:
  current level = [root]
  while current level is not empty:
   new level = []
   for every actor in the current level
    for every collaborator of the actor
        if the collaborator is not already in the tree
            add collaborator:actor to the tree
            add collaborator to the new level
   set current level = new level
*)

(* forEveryCollaborator : string -> string list -> string list -> Map.t -> (string list) * Map.t
 *
 * The call (forEveryCollaborator actor collaborators newLevel map) extends newLevel and map
 * by processing each collaborator c in collaborators as follows: if c isn't already a key in 
 * map then:
 * 1. the key/value mapping c -> actor is added to map and
 * 2. c is added to the newLevel.
 *)
let rec forEveryCollaborator actor collaborators newLevel map =
  match collaborators with
  | [] -> (newLevel, map)
  | collaborator :: collaborators ->
     (match M.mem collaborator map with
      | true  -> forEveryCollaborator actor collaborators newLevel map
      | false -> let map' = M.add collaborator actor map in
                 let newLevel' = collaborator :: newLevel
                 in
                 forEveryCollaborator actor collaborators newLevel' map')

(* forEveryActor : string list -> string list -> Map.t -> string list * Map.t
 *
 * The call (forEveryActor actors newLevel map) produces a newLevel and a new map by processing
 * the collaborators of each actor in actors.
 *)
let rec forEveryActor actors newLevel map =
  match actors with
  | [] -> (newLevel, map)
  | actor :: actors ->
     let collaborators = M.find actor coMap in
     let (newLevel', map') = forEveryCollaborator actor collaborators newLevel map
     in
     forEveryActor actors newLevel' map'

(* makeBFSTree : string -> Map.t
 *
 * The call (makeBFSTree actor) produces a map representing a breadth-first search of collaborators
 * of actor. The tree has key/value mappings of the form actor -> actor.  In particular, level 0 or
 * the root ofe the tree is the key/value mapping actor -> "". Then for each actor A at the current
 * level and for each collaborator C of A, if A isn't already a key in the map, then we add the 
 * key/value pair C -> A to the map and add C to the next level. 
 *)
let makeBFSTree actor =
  let rec repeat current tree =
    match current with
    | [] -> tree
    | _  -> let (newLevel', tree') = forEveryActor current [] tree
            in
            repeat newLevel' tree'
  in
  repeat [actor] (M.add actor "" M.empty)

(* The collaboration tree for Cate Blanchett.
 *)
let blanchettTree =
  let map = makeBFSTree "Cate Blanchett" in
  Printf.printf "\nBlanchett map has %d entries.\n" (M.cardinal map);
  map

let rec printPathFromTo actor root tree =
  match M.find actor tree with
  | "" -> Printf.printf "%s\n" root
  | actor' -> Printf.printf "%s\n" actor;
              printPathFromTo actor' root tree

let _ = printPathFromTo "Dustin Hoffman" "Cate Blanchett" blanchettTree

(* 14. Pick several other actors and print out the path from that
 *     actor to the root actor.  
 *)
let fourteen () =
  let root = "Morgan Freeman" in
  let map = makeBFSTree root
  in
  Printf.printf "\nFourteen:\n";
  printPathFromTo "Dileep Rao" root map

let _ = fourteen ()

(* 15. What percentage of the total number of actors in the database 
 * are in this tree?  (The answer is quite striking.)
 *)
let fifteen () =
  let percentage = (float (M.cardinal blanchettTree)) /. (float (M.cardinal afMap)) 
  in
  Printf.printf "\nFifteen: Percentage of Blanchett collaborators of all actors = %f.\n" percentage

let _ = fifteen()

(* 16. What is the length of the longest path in the tree? (To find 
 * this, you may need to revise your tree-making function ever so 
 * slightly so that it counts the number of levels, and prints out 
 * how many distinct levels it found.)
 *)
let sixteen () = ()
                   
(* 17. Find an instance of a longest path in the tree.  For instance, 
 * if you get the result that everyone on the tree is connected to 
 * Sandra Bullock by a path of at most 5, find an actor whose shortest 
 * path to Sandra Bullock has length 5, and print out the path.
 *)
let seventeen () = ()

(* PART 5 ************************************************)

